use io;

/* read instance data.*/
function input(){
    local usage = "Usage: localsolver clustered-vehicle-routing.lsp "
        + "inFileName=inputFile [lsTimeLimit=timeLimit]";

    if (inFileName == nil) throw usage;

    readInputCluvrp();
    computeDistanceMatrix();
    computeDistanceCustomerDepot();
}

/* Declare the optimization model */
function model() {
    // Sequences of clusters visited by each truck
    truckSequences[k in 0...nbTrucks] <- list(nbClusters);
    
    // A list is created for each cluster, to determine the order within the cluster
    for[k in 0...nbClusters] {
        clustersSequences[k] <- list(clusters[k].count());
        // All customers in the cluster must be visited 
        constraint count(clustersSequences[k]) == clusters[k].count();
    }
    
    // All clusters must be visited by the trucks
    constraint partition[k in 0...nbTrucks](truckSequences[k]);


    for[k in 0...nbClusters] {
        local sequence <- clustersSequences[k];
        local c <- count(sequence);

        // Distance traveled within cluster k
        clustersDistances[k] <- sum(1...c, i=> 
                distanceMatrix[clusters[k][sequence[i-1]]][clusters[k][sequence[i]]]
        );

        // first and last point visited when travelling into cluster k
        initialNodes[k] <- clusters[k][sequence[0]];
        endNodes[k] <- clusters[k][sequence[c-1]];
    }

    for [k in 0...nbTrucks] {
        local sequence <- truckSequences[k];
        local c <- count(sequence);

        // The quantity needed in each route (sum of quantity of each cluster) must not
        // exceed the truck capacity
        routeQuantity <- sum(sequence, j => demands[j]);
        constraint routeQuantity <= truckCapacity;

        // Distance traveled by truck k
        // = distance in each cluster + distance between cluster + distance with depot at 
        // the beginning end at the end of a route
        routeDistances[k] <- sum(1...c, i => clustersDistances[sequence[i]] 
                + distanceMatrix[endNodes[sequence[i - 1]]][initialNodes[sequence[i]]] ) 
                + (c > 0 ? clustersDistances[sequence[0]] 
                + distanceDepot[initialNodes[sequence[0]]] 
                + distanceDepot[endNodes[sequence[c-1]]] : 0) ;

            
    }
    // Objective:  minimize the distance traveled
    totalDistance <- sum[k in 0...nbTrucks](routeDistances[k]);
    minimize totalDistance;
}

/* Parametrize the solver */
function param() { 
    if (lsTimeLimit == nil) lsTimeLimit = 20;
}


/* Write the solution in a file with the following format:
 * - total distance
 * - for each truck the customers visited (omitting the start/end at the depot) */
function output() {
    if (solFileName == nil) return;
    local outfile = io.openWrite(solFileName);

    outfile.println(totalDistance.value);
    for [k in 0...nbTrucks] {
        for [cluster in truckSequences[k].value]{
            for [customer in clustersSequences[cluster].value]
                outfile.print(clusters[cluster][customer] + 2, " ");
        }
        outfile.println();
    }
    outfile.close();
}


function readInputCluvrp() {
    local inFile = io.openRead(inFileName);
    local nbNodes = 0;
    while (true) {
        local str = inFile.readString();
        if (str.startsWith("DIMENSION")) {
            if (!str.endsWith(":")) str = inFile.readString();
            nbNodes = inFile.readInt();
            nbCustomers = nbNodes - 1;
        } else if ((str.startsWith("VEHICLES"))) {
            if (!str.endsWith(":")) str = inFile.readString();
            nbTrucks = inFile.readInt();
        } else if ((str.startsWith("GVRP_SETS"))) {
            if (!str.endsWith(":")) str = inFile.readString();
            nbClusters = inFile.readInt();
        } else if ((str.startsWith("CAPACITY"))) {
            if (!str.endsWith(":")) str = inFile.readString();
            truckCapacity = inFile.readInt();
        } else if (str.startsWith("NODE_COORD_SECTION")) {
            break;
        } else {
            local dump = inFile.readln();
        }
    }
    for [n in 1..nbNodes] {
        if (n != inFile.readInt()) throw "Unexpected index";
        if(n==1){
            depotX =  round(inFile.readDouble());
            depotY =  round(inFile.readDouble());
        }else{
            // -2 because original customer indices are in 2..nbNodes
            customersX[n - 2] = round(inFile.readDouble());
            customersY[n - 2] = round(inFile.readDouble());
        }
    }

    dump = inFile.readln();
    if (!dump.startsWith("GVRP_SET_SECTION")) throw "Expected keyword GVRP_SET_SECTION";
    for [n in 1..nbClusters] {
        if (n != inFile.readInt()) throw "Unexpected index";
        value = inFile.readInt();
        i = 0;
        while(value != -1){
            // -2 because original customer indices are in 2..nbNodes
            clusters[n-1][i] = value - 2;
            value = inFile.readInt();
            i += 1;
        } 
    }

    dump = inFile.readln();
    if (!dump.startsWith("DEMAND_SECTION")) throw "Expected keyword DEMAND_SECTION";
    for [n in 1..nbClusters] {
        if (n != inFile.readInt()) throw "Unexpected index";
        demands[n-1] = inFile.readInt();
    }
}


/* Compute the distance matrix */
function computeDistanceMatrix() {
    for [i in 0...nbCustomers] {
        distanceMatrix[i][i] = 0;
        for [j in i+1...nbCustomers] {
            local localDistance = computeDist(i, j);
            distanceMatrix[j][i] = localDistance;
            distanceMatrix[i][j] = localDistance;
        }
    }
}

/* Compute the distance between customers and depot */
function computeDistanceCustomerDepot() {
    for [i in 0...nbCustomers] {
        distanceDepot[i] =  round(sqrt(pow((depotX - customersX[i]), 2)
                + pow((depotY - customersY[i]), 2)));
    }
}

function computeDist(i, j) {
    local exactDist = sqrt(pow((customersX[i] - customersX[j]), 2)
            + pow((customersY[i] - customersY[j]), 2));
    return round(exactDist);
}